@@ -2,10 +2,9 @@
  * @file JSArrayProxy.cc
  * @author Philippe Laporte (philippe@distributive.network)
  * @brief JSArrayProxy is a custom C-implemented python type that derives from list. It acts as a proxy for JSArrays from Spidermonkey, and behaves like a list would.
- * @version 0.1
  * @date 2023-11-22
  *
- * Copyright (c) 2023 Distributive Corp.
+ * @copyright Copyright (c) 2023-2024 Distributive Corp.
  *
  */
 
@@ -18,6 +17,7 @@
 #include "include/jsTypeFactory.hh"
 #include "include/pyTypeFactory.hh"
 #include "include/PyBaseProxyHandler.hh"
+#include "include/JSFunctionProxy.hh"
 
 #include <jsapi.h>
 #include <jsfriendapi.h>
@@ -27,25 +27,15 @@
 
 void JSArrayProxyMethodDefinitions::JSArrayProxy_dealloc(JSArrayProxy *self)
 {
-  self->jsArray.set(nullptr);
+  self->jsArray->set(nullptr);
+  delete self->jsArray;
   Py_TYPE(self)->tp_free((PyObject *)self);
 }
 
-PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
-{
-  return PyList_Type.tp_new(subtype, args, kwds);
-}
-
-int JSArrayProxyMethodDefinitions::JSArrayProxy_init(JSArrayProxy *self, PyObject *args, PyObject *kwds)
-{
-  PyList_Type.tp_init((PyObject *)self, args, kwds);
-  return 0;
-}
-
 Py_ssize_t JSArrayProxyMethodDefinitions::JSArrayProxy_length(JSArrayProxy *self)
 {
   uint32_t length;
-  JS::GetArrayLength(GLOBAL_CX, self->jsArray, &length);
+  JS::GetArrayLength(GLOBAL_CX, *(self->jsArray), &length);
   return (Py_ssize_t)length;
 }
 
@@ -53,30 +43,28 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get(JSArrayProxy *self, Py
 {
   JS::RootedId id(GLOBAL_CX);
   if (!keyToId(key, &id)) {
-    // TODO (Caleb Aikens): raise exception here
-    return NULL; // key is not a str or int
+    PyErr_SetString(PyExc_AttributeError, "JSArrayProxy property name must be of type str or int");
+    return NULL;
   }
 
   // look through the methods for dispatch and return key if no method found
   for (size_t index = 0;; index++) {
     const char *methodName = JSArrayProxyType.tp_methods[index].ml_name;
-    if (methodName == NULL) {   // reached end of list
-      JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
-      JS_GetPropertyById(GLOBAL_CX, self->jsArray, id, value);
-      JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
-      return pyTypeFactory(GLOBAL_CX, global, value)->getPyObject();
+    if (methodName == NULL || !PyUnicode_Check(key)) {   // reached end of list
+      JS::RootedValue value(GLOBAL_CX);
+      JS_GetPropertyById(GLOBAL_CX, *(self->jsArray), id, &value);
+      if (value.isUndefined() && PyUnicode_Check(key)) {
+        if (strcmp("__class__", PyUnicode_AsUTF8(key)) == 0) {
+          return PyObject_GenericGetAttr((PyObject *)self, key);
+        }
+      }
+      return pyTypeFactory(GLOBAL_CX, value)->getPyObject();
     }
-    else if (PyUnicode_Check(key)) {
+    else {
       if (strcmp(methodName, PyUnicode_AsUTF8(key)) == 0) {
         return PyObject_GenericGetAttr((PyObject *)self, key);
       }
     }
-    else {
-      JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
-      JS_GetPropertyById(GLOBAL_CX, self->jsArray, id, value);
-      JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
-      return pyTypeFactory(GLOBAL_CX, global, value)->getPyObject();
-    }
   }
 }
 
@@ -86,12 +74,12 @@ static PyObject *list_slice(JSArrayProxy *self, Py_ssize_t ilow, Py_ssize_t ihig
   JS::Rooted<JS::ValueArray<2>> jArgs(GLOBAL_CX);
   jArgs[0].setInt32(ilow);
   jArgs[1].setInt32(ihigh);
-  JS::RootedValue *jReturnedArray = new JS::RootedValue(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "slice", jArgs, jReturnedArray)) {
+  JS::RootedValue jReturnedArray(GLOBAL_CX);
+  if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "slice", jArgs, &jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jReturnedArray)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, jReturnedArray)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get_subscript(JSArrayProxy *self, PyObject *key)
@@ -116,10 +104,10 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get_subscript(JSArrayProxy
     JS::RootedId id(GLOBAL_CX);
     JS_IndexToId(GLOBAL_CX, index, &id);
 
-    JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
-    JS_GetPropertyById(GLOBAL_CX, self->jsArray, id, value);
+    JS::RootedValue value(GLOBAL_CX);
+    JS_GetPropertyById(GLOBAL_CX, *(self->jsArray), id, &value);
 
-    return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), value)->getPyObject();
+    return pyTypeFactory(GLOBAL_CX, value)->getPyObject();
   }
   else if (PySlice_Check(key)) {
     Py_ssize_t start, stop, step, slicelength, index;
@@ -137,18 +125,18 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get_subscript(JSArrayProxy
       return list_slice(self, start, stop);
     }
     else {
-      JS::RootedObject jCombinedArray = JS::RootedObject(GLOBAL_CX, JS::NewArrayObject(GLOBAL_CX, slicelength));
+      JS::RootedObject jCombinedArray(GLOBAL_CX, JS::NewArrayObject(GLOBAL_CX, slicelength));
 
       JS::RootedValue elementVal(GLOBAL_CX);
       for (size_t cur = start, index = 0; index < slicelength; cur += (size_t)step, index++) {
-        JS_GetElement(GLOBAL_CX, self->jsArray, cur, &elementVal);
+        JS_GetElement(GLOBAL_CX, *(self->jsArray), cur, &elementVal);
         JS_SetElement(GLOBAL_CX, jCombinedArray, index, elementVal);
       }
 
-      JS::RootedValue *jCombinedArrayValue = new JS::RootedValue(GLOBAL_CX);
-      jCombinedArrayValue->setObjectOrNull(jCombinedArray);
+      JS::RootedValue jCombinedArrayValue(GLOBAL_CX);
+      jCombinedArrayValue.setObjectOrNull(jCombinedArray);
 
-      return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jCombinedArrayValue)->getPyObject();
+      return pyTypeFactory(GLOBAL_CX, jCombinedArrayValue)->getPyObject();
     }
   }
   else {
@@ -223,37 +211,37 @@ static int list_ass_slice(JSArrayProxy *self, Py_ssize_t ilow, Py_ssize_t ihigh,
 
   if (selfLength + d == 0) {
     Py_XDECREF(v_as_SF);
-    JSArrayProxyMethodDefinitions::JSArrayProxy_clear(self);
+    JSArrayProxyMethodDefinitions::JSArrayProxy_clear_method(self);
     return 0;
   }
 
   if (d < 0) {   /* Delete -d items */
     JS::RootedValue elementVal(GLOBAL_CX);
     for (size_t index = ihigh, count = 0; count < selfLength - ihigh; index++, count++) {
-      JS_GetElement(GLOBAL_CX, self->jsArray, index, &elementVal);
-      JS_SetElement(GLOBAL_CX, self->jsArray, index+d, elementVal);
+      JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+      JS_SetElement(GLOBAL_CX, *(self->jsArray), index+d, elementVal);
     }
 
-    JS::SetArrayLength(GLOBAL_CX, self->jsArray, selfLength + d);
+    JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), selfLength + d);
   }
   else if (d > 0) { /* Insert d items */
     k = selfLength;
 
-    JS::SetArrayLength(GLOBAL_CX, self->jsArray, k + d);
+    JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), k + d);
 
     selfLength = k + d;
 
     JS::RootedValue elementVal(GLOBAL_CX);
     for (size_t index = ihigh, count = 0; count < k - ihigh; index++, count++) {
-      JS_GetElement(GLOBAL_CX, self->jsArray, index, &elementVal);
-      JS_SetElement(GLOBAL_CX, self->jsArray, index+d, elementVal);
+      JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+      JS_SetElement(GLOBAL_CX, *(self->jsArray), index+d, elementVal);
     }
   }
 
   JS::RootedValue elementVal(GLOBAL_CX);
   for (k = 0; k < n; k++, ilow++) {
     elementVal.set(jsTypeFactory(GLOBAL_CX, vitem[k]));
-    JS_SetElement(GLOBAL_CX, self->jsArray, ilow, elementVal);
+    JS_SetElement(GLOBAL_CX, *(self->jsArray), ilow, elementVal);
   }
 
   result = 0;
@@ -286,10 +274,10 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
 
     if (value) { // we are setting a value
       JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, value));
-      JS_SetPropertyById(GLOBAL_CX, self->jsArray, id, jValue);
+      JS_SetPropertyById(GLOBAL_CX, *(self->jsArray), id, jValue);
     } else { // we are deleting a value
       JS::ObjectOpResult ignoredResult;
-      JS_DeletePropertyById(GLOBAL_CX, self->jsArray, id, ignoredResult);
+      JS_DeletePropertyById(GLOBAL_CX, *(self->jsArray), id, ignoredResult);
     }
 
     return 0;
@@ -345,8 +333,8 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
         }
 
         for (size_t index = cur, count = 0; count < lim; index++, count++) {
-          JS_GetElement(GLOBAL_CX, self->jsArray, index + 1, &elementVal);
-          JS_SetElement(GLOBAL_CX, self->jsArray, index - i, elementVal);
+          JS_GetElement(GLOBAL_CX, *(self->jsArray), index + 1, &elementVal);
+          JS_SetElement(GLOBAL_CX, *(self->jsArray), index - i, elementVal);
         }
       }
 
@@ -354,12 +342,12 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
 
       if (cur < (size_t)selfSize) {
         for (size_t index = cur, count = 0; count < selfSize - cur; index++, count++) {
-          JS_GetElement(GLOBAL_CX, self->jsArray, index, &elementVal);
-          JS_SetElement(GLOBAL_CX, self->jsArray, index - slicelength, elementVal);
+          JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+          JS_SetElement(GLOBAL_CX, *(self->jsArray), index - slicelength, elementVal);
         }
       }
 
-      JS::SetArrayLength(GLOBAL_CX, self->jsArray, selfSize - slicelength);
+      JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), selfSize - slicelength);
 
       return 0;
     }
@@ -399,7 +387,7 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
       JS::RootedValue elementVal(GLOBAL_CX);
       for (cur = start, i = 0; i < slicelength; cur += (size_t)step, i++) {
         elementVal.set(jsTypeFactory(GLOBAL_CX, seqitems[i]));
-        JS_SetElement(GLOBAL_CX, self->jsArray, cur, elementVal);
+        JS_SetElement(GLOBAL_CX, *(self->jsArray), cur, elementVal);
       }
 
       Py_DECREF(seq);
@@ -447,20 +435,18 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_richcompare(JSArrayProxy *
     }
   }
 
-  JS::RootedValue *elementVal;
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
+  JS::RootedValue elementVal(GLOBAL_CX);
 
   Py_ssize_t index;
   /* Search for the first index where items are different */
   for (index = 0; index < selfLength && index < otherLength; index++) {
-    elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
 
-    PyObject *leftItem = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+    PyObject *leftItem = pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
     PyObject *rightItem;
     if (PyObject_TypeCheck(other, &JSArrayProxyType)) {
-      JS_GetElement(GLOBAL_CX, ((JSArrayProxy *)other)->jsArray, index, elementVal);
-      rightItem = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+      JS_GetElement(GLOBAL_CX, *(((JSArrayProxy *)other)->jsArray), index, &elementVal);
+      rightItem = pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
     } else {
       rightItem = ((PyListObject *)other)->ob_item[index];
     }
@@ -495,10 +481,9 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_richcompare(JSArrayProxy *
     Py_RETURN_TRUE;
   }
 
-  elementVal = new JS::RootedValue(GLOBAL_CX);
-  JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
+  JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
   /* Compare the final item again using the proper operator */
-  return PyObject_RichCompare(pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject(), ((PyListObject *)other)->ob_item[index], op);
+  return PyObject_RichCompare(pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject(), ((PyListObject *)other)->ob_item[index], op);
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repr(JSArrayProxy *self) {
@@ -520,7 +505,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repr(JSArrayProxy *self) {
   /* "[" + "1" + ", 2" * (len - 1) + "]" */
   writer.min_length = 1 + 1 + (2 + 1) * (selfLength - 1) + 1;
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
+  JS::RootedValue elementVal(GLOBAL_CX);
 
   if (_PyUnicodeWriter_WriteChar(&writer, '[') < 0) {
     goto error;
@@ -534,14 +519,13 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repr(JSArrayProxy *self) {
       }
     }
 
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
 
     PyObject *s;
-    if (&elementVal->toObject() == self->jsArray.get()) {
+    if (&elementVal.toObject() == (*(self->jsArray)).get()) {
       s = PyObject_Repr((PyObject *)self);
     } else {
-      s = PyObject_Repr(pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject());
+      s = PyObject_Repr(pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject());
     }
     if (s == NULL) {
       goto error;
@@ -621,18 +605,18 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_concat(JSArrayProxy *self,
     }
   }
 
-  JS::RootedObject jCombinedArray = JS::RootedObject(GLOBAL_CX, JS::NewArrayObject(GLOBAL_CX, (size_t)sizeSelf + (size_t)sizeValue));
+  JS::RootedObject jCombinedArray(GLOBAL_CX, JS::NewArrayObject(GLOBAL_CX, (size_t)sizeSelf + (size_t)sizeValue));
 
   JS::RootedValue elementVal(GLOBAL_CX);
 
   for (Py_ssize_t inputIdx = 0; inputIdx < sizeSelf; inputIdx++) {
-    JS_GetElement(GLOBAL_CX, self->jsArray, inputIdx, &elementVal);
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), inputIdx, &elementVal);
     JS_SetElement(GLOBAL_CX, jCombinedArray, inputIdx, elementVal);
   }
 
   if (PyObject_TypeCheck(value, &JSArrayProxyType)) {
     for (Py_ssize_t inputIdx = 0; inputIdx < sizeValue; inputIdx++) {
-      JS_GetElement(GLOBAL_CX, ((JSArrayProxy *)value)->jsArray, inputIdx, &elementVal);
+      JS_GetElement(GLOBAL_CX, *(((JSArrayProxy *)value)->jsArray), inputIdx, &elementVal);
       JS_SetElement(GLOBAL_CX, jCombinedArray, sizeSelf + inputIdx, elementVal);
     }
   } else {
@@ -643,10 +627,10 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_concat(JSArrayProxy *self,
     }
   }
 
-  JS::RootedValue *jCombinedArrayValue = new JS::RootedValue(GLOBAL_CX);
-  jCombinedArrayValue->setObjectOrNull(jCombinedArray);
+  JS::RootedValue jCombinedArrayValue(GLOBAL_CX);
+  jCombinedArrayValue.setObjectOrNull(jCombinedArray);
 
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jCombinedArrayValue)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, jCombinedArrayValue)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repeat(JSArrayProxy *self, Py_ssize_t n) {
@@ -659,12 +643,12 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repeat(JSArrayProxy *self,
     return PyErr_NoMemory();
   }
 
-  JS::RootedObject jCombinedArray = JS::RootedObject(GLOBAL_CX, JS::NewArrayObject(GLOBAL_CX, input_size * n));
+  JS::RootedObject jCombinedArray(GLOBAL_CX, JS::NewArrayObject(GLOBAL_CX, input_size * n));
   // repeat within new array
   // one might think of using copyWithin but in SpiderMonkey it's implemented in JS!
   JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t inputIdx = 0; inputIdx < input_size; inputIdx++) {
-    JS_GetElement(GLOBAL_CX, self->jsArray, inputIdx, &elementVal);
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), inputIdx, &elementVal);
     for (Py_ssize_t repeatIdx = 0; repeatIdx < n; repeatIdx++) {
       JS_SetElement(GLOBAL_CX, jCombinedArray, repeatIdx * input_size + inputIdx, elementVal);
     }
@@ -673,7 +657,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repeat(JSArrayProxy *self,
   JS::RootedValue jCombinedArrayValue(GLOBAL_CX);
   jCombinedArrayValue.setObjectOrNull(jCombinedArray);
 
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), &jCombinedArrayValue)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, jCombinedArrayValue)->getPyObject();
 }
 
 int JSArrayProxyMethodDefinitions::JSArrayProxy_contains(JSArrayProxy *self, PyObject *element) {
@@ -681,11 +665,11 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_contains(JSArrayProxy *self, PyO
   int cmp;
 
   Py_ssize_t numElements = JSArrayProxy_length(self);
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
+
+  JS::RootedValue elementVal(GLOBAL_CX);
   for (index = 0, cmp = 0; cmp == 0 && index < numElements; ++index) {
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
-    PyObject *item = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+    PyObject *item = pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
     Py_INCREF(item);
     cmp = PyObject_RichCompareBool(item, element, Py_EQ);
     Py_DECREF(item);
@@ -698,7 +682,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_concat(JSArrayProx
   Py_ssize_t valueLength = Py_SIZE(value);
 
   // allocate extra spacePy_SIZE
-  JS::SetArrayLength(GLOBAL_CX, self->jsArray, selfLength + valueLength);
+  JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), selfLength + valueLength);
 
   JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, value));
   JS::RootedObject jRootedValue = JS::RootedObject(GLOBAL_CX, jValue.toObjectOrNull());
@@ -706,7 +690,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_concat(JSArrayProx
   JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t inputIdx = 0; inputIdx < valueLength; inputIdx++) {
     JS_GetElement(GLOBAL_CX, jRootedValue, inputIdx, &elementVal);
-    JS_SetElement(GLOBAL_CX, self->jsArray, selfLength + inputIdx, elementVal);
+    JS_SetElement(GLOBAL_CX, *(self->jsArray), selfLength + inputIdx, elementVal);
   }
 
   Py_INCREF(self);
@@ -721,7 +705,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_repeat(JSArrayProx
   }
 
   if (n < 1) {
-    JSArrayProxy_clear(self);
+    JSArrayProxy_clear_method(self);
     Py_INCREF(self);
     return (PyObject *)self;
   }
@@ -730,15 +714,15 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_repeat(JSArrayProx
     return PyErr_NoMemory();
   }
 
-  JS::SetArrayLength(GLOBAL_CX, self->jsArray, input_size * n);
+  JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), input_size * n);
 
   // repeat within self
   // one might think of using copyWithin but in SpiderMonkey it's implemented in JS!
   JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t inputIdx = 0; inputIdx < input_size; inputIdx++) {
-    JS_GetElement(GLOBAL_CX, self->jsArray, inputIdx, &elementVal);
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), inputIdx, &elementVal);
     for (Py_ssize_t repeatIdx = 0; repeatIdx < n; repeatIdx++) {
-      JS_SetElement(GLOBAL_CX, self->jsArray, repeatIdx * input_size + inputIdx, elementVal);
+      JS_SetElement(GLOBAL_CX, *(self->jsArray), repeatIdx * input_size + inputIdx, elementVal);
     }
   }
 
@@ -746,44 +730,29 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_repeat(JSArrayProx
   return (PyObject *)self;
 }
 
-PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_clear(JSArrayProxy *self) {
-  JS::SetArrayLength(GLOBAL_CX, self->jsArray, 0);
+PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_clear_method(JSArrayProxy *self) {
+  JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), 0);
   Py_RETURN_NONE;
 }
 
-int JSArrayProxyMethodDefinitions::JSArrayProxy_clear_slot(JSArrayProxy *self) {
-  JSArrayProxyMethodDefinitions::JSArrayProxy_clear(self);
-  return 0;
-}
-
-int JSArrayProxyMethodDefinitions::JSArrayProxy_traverse(JSArrayProxy *self, visitproc visit, void *arg) {
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
-  for (Py_ssize_t i = JSArrayProxy_length(self); --i >= 0; ) {
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, i, elementVal);
-    Py_VISIT(pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject());
-  }
-  return 0;
-}
-
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_copy(JSArrayProxy *self) {
   JS::Rooted<JS::ValueArray<2>> jArgs(GLOBAL_CX);
   jArgs[0].setInt32(0);
   jArgs[1].setInt32(JSArrayProxy_length(self));
-  JS::RootedValue *jReturnedArray = new JS::RootedValue(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "slice", jArgs, jReturnedArray)) {
+  JS::RootedValue jReturnedArray(GLOBAL_CX);
+  if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "slice", jArgs, &jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jReturnedArray)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, jReturnedArray)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_append(JSArrayProxy *self, PyObject *value) {
   Py_ssize_t len = JSArrayProxy_length(self);
 
-  JS::SetArrayLength(GLOBAL_CX, self->jsArray, len + 1);
+  JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), len + 1);
   JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, value));
-  JS_SetElement(GLOBAL_CX, self->jsArray, len, jValue);
+  JS_SetElement(GLOBAL_CX, *(self->jsArray), len, jValue);
 
   Py_RETURN_NONE;
 }
@@ -830,7 +799,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_insert(JSArrayProxy *self,
   jArgs[2].set(jsTypeFactory(GLOBAL_CX, value));
 
   JS::RootedValue jReturnedArray(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "splice", jArgs, &jReturnedArray)) {
+  if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "splice", jArgs, &jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
@@ -853,14 +822,14 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_extend(JSArrayProxy *self,
 
     Py_ssize_t m = JSArrayProxy_length(self);
 
-    JS::SetArrayLength(GLOBAL_CX, self->jsArray, m + n);
+    JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), m + n);
 
     // populate the end of self with iterable's items.
     PyObject **src = PySequence_Fast_ITEMS(iterable);
     for (Py_ssize_t i = 0; i < n; i++) {
       PyObject *o = src[i];
       JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, o));
-      JS_SetElement(GLOBAL_CX, self->jsArray, m + i, jValue);
+      JS_SetElement(GLOBAL_CX, *(self->jsArray), m + i, jValue);
     }
 
     Py_DECREF(iterable);
@@ -889,9 +858,9 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_extend(JSArrayProxy *self,
         break;
       }
 
-      JS::SetArrayLength(GLOBAL_CX, self->jsArray, len + 1);
+      JS::SetArrayLength(GLOBAL_CX, *(self->jsArray), len + 1);
       JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, item));
-      JS_SetElement(GLOBAL_CX, self->jsArray, len, jValue);
+      JS_SetElement(GLOBAL_CX, *(self->jsArray), len, jValue);
       len++;
     }
 
@@ -942,27 +911,26 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_pop(JSArrayProxy *self, Py
   jArgs[1].setInt32(1);
 
   JS::RootedValue jReturnedArray(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "splice", jArgs, &jReturnedArray)) {
+  if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "splice", jArgs, &jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
 
   // need the value in the returned array, not the array itself
   JS::RootedObject rootedReturnedArray(GLOBAL_CX, jReturnedArray.toObjectOrNull());
-  JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-  JS_GetElement(GLOBAL_CX, rootedReturnedArray, 0, elementVal);
+  JS::RootedValue elementVal(GLOBAL_CX);
+  JS_GetElement(GLOBAL_CX, rootedReturnedArray, 0, &elementVal);
 
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), elementVal)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_remove(JSArrayProxy *self, PyObject *value) {
   Py_ssize_t selfSize = JSArrayProxy_length(self);
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
+  JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t index = 0; index < selfSize; index++) {
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
-    PyObject *obj = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+    PyObject *obj = pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
     Py_INCREF(obj);
     int cmp = PyObject_RichCompareBool(obj, value, Py_EQ);
     Py_DECREF(obj);
@@ -971,7 +939,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_remove(JSArrayProxy *self,
       jArgs[0].setInt32(index);
       jArgs[1].setInt32(1);
       JS::RootedValue jReturnedArray(GLOBAL_CX);
-      if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "splice", jArgs, &jReturnedArray)) {
+      if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "splice", jArgs, &jReturnedArray)) {
         PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
         return NULL;
       }
@@ -1024,11 +992,10 @@ skip_optional:
     }
   }
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
+  JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t index = start; index < stop && index < selfSize; index++) {
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
-    PyObject *obj = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+    PyObject *obj = pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
     Py_INCREF(obj);
     int cmp = PyObject_RichCompareBool(obj, value, Py_EQ);
     Py_DECREF(obj);
@@ -1048,11 +1015,10 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_count(JSArrayProxy *self,
   Py_ssize_t count = 0;
 
   Py_ssize_t length = JSArrayProxy_length(self);
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
+  JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t index = 0; index < length; index++) {
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
-    PyObject *obj = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+    JS_GetElement(GLOBAL_CX, *(self->jsArray), index, &elementVal);
+    PyObject *obj = pyTypeFactory(GLOBAL_CX, elementVal)->getPyObject();
     Py_INCREF(obj);
     int cmp = PyObject_RichCompareBool(obj, value, Py_EQ);
     Py_DECREF(obj);
@@ -1069,7 +1035,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_count(JSArrayProxy *self,
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_reverse(JSArrayProxy *self) {
   if (JSArrayProxy_length(self) > 1) {
     JS::RootedValue jReturnedArray(GLOBAL_CX);
-    if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "reverse", JS::HandleValueArray::empty(), &jReturnedArray)) {
+    if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "reverse", JS::HandleValueArray::empty(), &jReturnedArray)) {
       PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
       return NULL;
     }
@@ -1098,18 +1064,15 @@ static bool sort_compare_key_func(JSContext *cx, unsigned argc, JS::Value *vp) {
   }
   bool reverse = reverseValue.toBoolean();
 
-
-  JS::RootedObject *global = new JS::RootedObject(cx, JS::GetNonCCWObjectGlobal(&args.callee()));
-
-  JS::RootedValue *elementVal = new JS::RootedValue(cx, args[0]);
-  PyObject *args_0 = pyTypeFactory(cx, global, elementVal)->getPyObject();
+  JS::RootedValue elementVal0(cx, args[0]);
+  PyObject *args_0 = pyTypeFactory(cx, elementVal0)->getPyObject();
   PyObject *args_0_result = PyObject_CallFunction(keyfunc, "O", args_0);
   if (!args_0_result) {
     return false;
   }
 
-  elementVal = new JS::RootedValue(cx, args[1]);
-  PyObject *args_1 = pyTypeFactory(cx, global, elementVal)->getPyObject();
+  JS::RootedValue elementVal1(cx, args[1]);
+  PyObject *args_1 = pyTypeFactory(cx, elementVal1)->getPyObject();
   PyObject *args_1_result = PyObject_CallFunction(keyfunc, "O", args_1);
   if (!args_1_result) {
     return false;
@@ -1149,13 +1112,11 @@ static bool sort_compare_default(JSContext *cx, unsigned argc, JS::Value *vp) {
   }
   bool reverse = reverseValue.toBoolean();
 
-  JS::RootedObject *global = new JS::RootedObject(cx, JS::GetNonCCWObjectGlobal(&args.callee()));
+  JS::RootedValue elementVal0(cx, args[0]);
+  PyObject *args_0 = pyTypeFactory(cx, elementVal0)->getPyObject();
 
-  JS::RootedValue *elementVal = new JS::RootedValue(cx, args[0]);
-  PyObject *args_0 = pyTypeFactory(cx, global, elementVal)->getPyObject();
-
-  elementVal = new JS::RootedValue(cx, args[1]);
-  PyObject *args_1 = pyTypeFactory(cx, global, elementVal)->getPyObject();
+  JS::RootedValue elementVal1(cx, args[1]);
+  PyObject *args_1 = pyTypeFactory(cx, elementVal1)->getPyObject();
 
   int cmp = PyObject_RichCompareBool(args_0, args_1, Py_LT);
   if (cmp > 0) {
@@ -1259,7 +1220,7 @@ skip_optional_kwonly:
 
           JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
           jArgs[0].setObject(*funObj);
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
+          if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "sort", jArgs, &jReturnedArray)) {
             if (!PyErr_Occurred()) {
               PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
             }
@@ -1281,7 +1242,7 @@ skip_optional_kwonly:
           // two-arg js-style
           JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
           jArgs[0].set(jsTypeFactory(GLOBAL_CX, keyfunc));
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
+          if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "sort", jArgs, &jReturnedArray)) {
             PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
             return NULL;
           }
@@ -1291,48 +1252,41 @@ skip_optional_kwonly:
           }
         }
       }
-      else if (PyCFunction_Check(keyfunc)) {
-        // check if builtin 1-arg python or js 2-arg compare
-        int flags = PyCFunction_GetFlags((PyObject *)keyfunc);
-
-        if (flags & METH_VARARGS && !(flags & METH_KEYWORDS)) {
-          // we got a JS compare function, use it as-is
-          JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
-          jArgs[0].set(jsTypeFactory(GLOBAL_CX, keyfunc));
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
-            PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
-            return NULL;
-          }
-
-          if (reverse) {
-            JSArrayProxy_reverse(self);
-          }
+      else if (PyObject_TypeCheck(keyfunc, &JSFunctionProxyType)) {
+        JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
+        jArgs[0].setObject(**((JSFunctionProxy *)keyfunc)->jsFunc);
+        if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "sort", jArgs, &jReturnedArray)) {
+          PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
+          return NULL;
         }
-        else {
-          // we got a built-in python function
-          JS::RootedObject funObj(GLOBAL_CX, JS_GetFunctionObject(JS_NewFunction(GLOBAL_CX, sort_compare_key_func, 2, 0, NULL)));
 
-          JS::RootedValue privateValue(GLOBAL_CX, JS::PrivateValue(keyfunc));
-          if (!JS_SetProperty(GLOBAL_CX, funObj, "_key_func_param", privateValue)) {  // JS::SetReservedSlot(functionObj, KeyFuncSlot, JS::PrivateValue(keyfunc)); does not work
+        if (reverse) {
+          JSArrayProxy_reverse(self);
+        }
+      }
+      else if (PyCFunction_Check(keyfunc)) {
+        JS::RootedObject funObj(GLOBAL_CX, JS_GetFunctionObject(JS_NewFunction(GLOBAL_CX, sort_compare_key_func, 2, 0, NULL)));
+
+        JS::RootedValue privateValue(GLOBAL_CX, JS::PrivateValue(keyfunc));
+        if (!JS_SetProperty(GLOBAL_CX, funObj, "_key_func_param", privateValue)) {  // JS::SetReservedSlot(functionObj, KeyFuncSlot, JS::PrivateValue(keyfunc)); does not work
+          PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
+          return NULL;
+        }
+
+        JS::RootedValue reverseValue(GLOBAL_CX);
+        reverseValue.setBoolean(reverse);
+        if (!JS_SetProperty(GLOBAL_CX, funObj, "_reverse_param", reverseValue)) {
+          PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
+          return NULL;
+        }
+
+        JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
+        jArgs[0].setObject(*funObj);
+        if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "sort", jArgs, &jReturnedArray)) {
+          if (!PyErr_Occurred()) {
             PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
-            return NULL;
-          }
-
-          JS::RootedValue reverseValue(GLOBAL_CX);
-          reverseValue.setBoolean(reverse);
-          if (!JS_SetProperty(GLOBAL_CX, funObj, "_reverse_param", reverseValue)) {
-            PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
-            return NULL;
-          }
-
-          JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
-          jArgs[0].setObject(*funObj);
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
-            if (!PyErr_Occurred()) {
-              PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
-            }
-            return NULL;
           }
+          return NULL;
         }
       }
       else {
@@ -1354,7 +1308,7 @@ skip_optional_kwonly:
 
       JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
       jArgs[0].setObject(*funObj);
-      if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
+      if (!JS_CallFunctionName(GLOBAL_CX, *(self->jsArray), "sort", jArgs, &jReturnedArray)) {
         PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
         return NULL;
       }
