@@ -2,17 +2,14 @@
  * @file PromiseType.cc
  * @author Tom Tang (xmader@distributive.network)
  * @brief Struct for representing Promises
- * @version 0.1
  * @date 2023-03-29
  *
- * @copyright Copyright (c) 2023
+ * @copyright Copyright (c) 2023 Distributive Corp.
  *
  */
 
 #include "include/modules/pythonmonkey/pythonmonkey.hh"
-
 #include "include/PromiseType.hh"
-
 #include "include/PyEventLoop.hh"
 #include "include/PyType.hh"
 #include "include/TypeEnum.hh"
@@ -25,25 +22,22 @@
 
 #include <Python.h>
 
-#define PY_FUTURE_OBJ_SLOT 0 // slot id to access the python object in JS callbacks
+// slot ids to access the python object in JS callbacks
+#define PY_FUTURE_OBJ_SLOT 0
 #define PROMISE_OBJ_SLOT 1
-// slot id must be less than 2 (https://hg.mozilla.org/releases/mozilla-esr102/file/tip/js/src/vm/JSFunction.h#l866), otherwise it will access to arbitrary unsafe memory locations
 
 static bool onResolvedCb(JSContext *cx, unsigned argc, JS::Value *vp) {
   JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
 
   // Get the Promise state
   JS::Value promiseObjVal = js::GetFunctionNativeReserved(&args.callee(), PROMISE_OBJ_SLOT);
-  JS::RootedObject promise = JS::RootedObject(cx, &promiseObjVal.toObject());
+  JS::RootedObject promise(cx, &promiseObjVal.toObject());
   JS::PromiseState state = JS::GetPromiseState(promise);
 
   // Convert the Promise's result (either fulfilled resolution or rejection reason) to a Python object
-  // FIXME (Tom Tang): memory leak, not free-ed
   //  The result might be another JS function, so we must keep them alive
-  JS::RootedObject *thisv = new JS::RootedObject(cx);
-  args.computeThis(cx, thisv); // thisv is the global object, not the promise
-  JS::RootedValue *resultArg = new JS::RootedValue(cx, args[0]);
-  PyObject *result = pyTypeFactory(cx, thisv, resultArg)->getPyObject();
+  JS::RootedValue resultArg(cx, args[0]);
+  PyObject *result = pyTypeFactory(cx, resultArg)->getPyObject();
   if (state == JS::PromiseState::Rejected && !PyExceptionInstance_Check(result)) {
     // Wrap the result object into a SpiderMonkeyError object
     // because only *Exception objects can be thrown in Python `raise` statement and alike
@@ -81,7 +75,7 @@ PromiseType::PromiseType(JSContext *cx, JS::HandleObject promise) {
 
   // Callbacks to settle the Python asyncio.Future once the JS Promise is resolved
   JS::RootedObject onResolved = JS::RootedObject(cx, (JSObject *)js::NewFunctionWithReserved(cx, onResolvedCb, 1, 0, NULL));
-  js::SetFunctionNativeReserved(onResolved, PY_FUTURE_OBJ_SLOT, JS::PrivateValue(future.getFutureObject())); // put the address of the Python object in private slot so we can access it later
+  js::SetFunctionNativeReserved(onResolved, PY_FUTURE_OBJ_SLOT, JS::PrivateValue(future.getFutureObject()));
   js::SetFunctionNativeReserved(onResolved, PROMISE_OBJ_SLOT, JS::ObjectValue(*promise));
   AddPromiseReactions(cx, promise, onResolved, onResolved);
 
@@ -91,7 +85,7 @@ PromiseType::PromiseType(JSContext *cx, JS::HandleObject promise) {
 // Callback to resolve or reject the JS Promise when the Future is done
 static PyObject *futureOnDoneCallback(PyObject *futureCallbackTuple, PyObject *args) {
   JSContext *cx = (JSContext *)PyLong_AsVoidPtr(PyTuple_GetItem(futureCallbackTuple, 0));
-  auto rootedPtr = (JS::PersistentRooted<JSObject *> *)PyLong_AsVoidPtr(PyTuple_GetItem(futureCallbackTuple, 1));
+  JS::PersistentRootedObject *rootedPtr = (JS::PersistentRootedObject *)PyLong_AsVoidPtr(PyTuple_GetItem(futureCallbackTuple, 1));
   JS::HandleObject promise = *rootedPtr;
   PyObject *futureObj = PyTuple_GetItem(args, 0); // the callback is called with the Future object as its only argument
                                                   // see https://docs.python.org/3.9/library/asyncio-future.html#asyncio.Future.add_done_callback
