@@ -1,3 +1,14 @@
+/**
+ * @file PyEventLoop.cc
+ * @author Tom Tang (xmader@distributive.network)
+ * @brief Send jobs to the Python event-loop
+ * @date 2023-04-05
+ *
+ * @copyright Copyright (c) 2023 Distributive Corp.
+ *
+ */
+
+
 #include "include/PyEventLoop.hh"
 
 #include <Python.h>
@@ -6,7 +17,7 @@
  * @brief Wrapper to decrement the counter of queueing event-loop jobs after the job finishes
  */
 static PyObject *eventLoopJobWrapper(PyObject *jobFn, PyObject *Py_UNUSED(_)) {
-  PyObject *ret = PyObject_CallObject(jobFn, NULL); // jobFn()
+  PyObject *ret = PyObject_CallObject(jobFn, NULL);
   Py_XDECREF(ret); // don't care about its return value
   PyEventLoop::_locker->decCounter();
   if (PyErr_Occurred()) {
@@ -15,27 +26,47 @@ static PyObject *eventLoopJobWrapper(PyObject *jobFn, PyObject *Py_UNUSED(_)) {
     Py_RETURN_NONE;
   }
 }
-static PyMethodDef jobWrapperDef = {"eventLoopJobWrapper", eventLoopJobWrapper, METH_NOARGS, NULL};
+static PyMethodDef loopJobWrapperDef = {"eventLoopJobWrapper", eventLoopJobWrapper, METH_NOARGS, NULL};
+
+/**
+ * @brief Wrapper to remove the reference of the timer after the job finishes
+ */
+static PyObject *timerJobWrapper(PyObject *jobFn, PyObject *handlerPtr) {
+  auto handle = (PyEventLoop::AsyncHandle *)PyLong_AsVoidPtr(handlerPtr);
+  PyObject *ret = PyObject_CallObject(jobFn, NULL); // jobFn()
+  Py_XDECREF(ret); // don't care about its return value
+  handle->removeRef();
+  if (PyErr_Occurred()) {
+    return NULL;
+  } else {
+    Py_RETURN_NONE;
+  }
+}
+static PyMethodDef timerJobWrapperDef = {"timerJobWrapper", timerJobWrapper, METH_O, NULL};
 
 PyEventLoop::AsyncHandle PyEventLoop::enqueue(PyObject *jobFn) {
   PyEventLoop::_locker->incCounter();
-  PyObject *wrapper = PyCFunction_New(&jobWrapperDef, jobFn);
+  PyObject *wrapper = PyCFunction_New(&loopJobWrapperDef, jobFn);
   // Enqueue job to the Python event-loop
   //    https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.call_soon
-  PyObject *asyncHandle = PyObject_CallMethod(_loop, "call_soon_threadsafe", "O", wrapper); // https://docs.python.org/3/c-api/arg.html#c.Py_BuildValue
+  PyObject *asyncHandle = PyObject_CallMethod(_loop, "call_soon_threadsafe", "O", wrapper);
   return PyEventLoop::AsyncHandle(asyncHandle);
 }
 
-PyEventLoop::AsyncHandle PyEventLoop::enqueueWithDelay(PyObject *jobFn, double delaySeconds) {
-  PyEventLoop::_locker->incCounter();
-  PyObject *wrapper = PyCFunction_New(&jobWrapperDef, jobFn);
+PyEventLoop::AsyncHandle::id_ptr_pair PyEventLoop::enqueueWithDelay(PyObject *jobFn, double delaySeconds) {
+  auto handler = PyEventLoop::AsyncHandle::newEmpty();
+  PyObject *wrapper = PyCFunction_New(&timerJobWrapperDef, jobFn);
+  PyObject *handlerPtr = PyLong_FromVoidPtr(handler.second);
   // Schedule job to the Python event-loop
   //    https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.call_later
-  PyObject *asyncHandle = PyObject_CallMethod(_loop, "call_later", "dO", delaySeconds, wrapper); // https://docs.python.org/3/c-api/arg.html#c.Py_BuildValue
+  PyObject *asyncHandle = PyObject_CallMethod(_loop, "call_later", "dOO", delaySeconds, wrapper, handlerPtr); // https://docs.python.org/3/c-api/arg.html#c.Py_BuildValue
   if (asyncHandle == nullptr) {
     PyErr_Print(); // RuntimeError: Non-thread-safe operation invoked on an event loop other than the current one
+    return handler;
   }
-  return PyEventLoop::AsyncHandle(asyncHandle);
+  handler.second->swap(asyncHandle);
+  handler.second->addRef();
+  return handler;
 }
 
 PyEventLoop::Future PyEventLoop::createFuture() {
@@ -116,7 +147,7 @@ PyThreadState *PyEventLoop::_getMainThread() {
   // The last element in the linked-list of threads associated with the main interpreter should be the main thread
   // (The first element is the current thread, see https://github.com/python/cpython/blob/7cb3a44/Python/pystate.c#L291-L293)
   PyInterpreterState *interp = PyInterpreterState_Main();
-  PyThreadState *tstate = PyInterpreterState_ThreadHead(interp); // https://docs.python.org/3/c-api/init.html#c.PyInterpreterState_ThreadHead
+  PyThreadState *tstate = PyInterpreterState_ThreadHead(interp);
   while (PyThreadState_Next(tstate) != nullptr) {
     tstate = PyThreadState_Next(tstate);
   }
@@ -127,7 +158,7 @@ PyThreadState *PyEventLoop::_getMainThread() {
 PyThreadState *PyEventLoop::_getCurrentThread() {
   // `PyThreadState_Get` is used under the hood of the Python `asyncio.get_running_loop` method,
   // see https://github.com/python/cpython/blob/7cb3a44/Modules/_asynciomodule.c#L234
-  return PyThreadState_Get(); // https://docs.python.org/3/c-api/init.html#c.PyThreadState_Get
+  return PyThreadState_Get();
 }
 
 /* static */
@@ -145,7 +176,7 @@ void PyEventLoop::AsyncHandle::cancel() {
                                                                        // NULL if no such attribute (on a strict asyncio.Handle returned by loop.call_soon)
   bool finishedOrCanceled = scheduled && scheduled == Py_False; // the job function has already been executed or canceled
   if (!finishedOrCanceled) {
-    PyEventLoop::_locker->decCounter();
+    removeRef(); // automatically unref at finish
   }
   Py_XDECREF(scheduled);
 
