@@ -2,6 +2,21 @@ import pytest
 import pythonmonkey as pm
 import asyncio
 
+def test_timers_unref():
+    async def async_fn():
+        obj = {'val': 0}
+        pm.eval("""(obj) => {
+            setTimeout(()=>{ obj.val = 2 }, 1000).ref().ref().unref().ref().unref().unref(); 
+                // chaining, no use on the first two ref calls since it's already refed initially
+            setTimeout(()=>{ obj.val = 1 }, 100);
+        }""")(obj)
+        await pm.wait() # we shouldn't wait until the first timer is fired since it's currently unrefed
+        assert obj['val'] == 1
+
+        # making sure the async_fn is run
+        return True
+    assert asyncio.run(async_fn())
+
 def test_set_clear_timeout():
     # throw RuntimeError outside a coroutine
     with pytest.raises(RuntimeError, match="PythonMonkey cannot find a running Python event-loop to make asynchronous calls."):
@@ -21,9 +36,10 @@ def test_set_clear_timeout():
         def to_raise(msg):
             f1.set_exception(TypeError(msg))
         timeout_id0 = pm.eval("setTimeout")(to_raise, 100, "going to be there")
-        assert type(timeout_id0) == float
-        assert timeout_id0 > 0                  # `setTimeout` should return a positive integer value
-        assert int(timeout_id0) == timeout_id0
+        # `setTimeout` should return a `Timeout` instance wrapping a positive integer value
+        assert pm.eval("(t) => t instanceof setTimeout.Timeout")(timeout_id0)
+        assert pm.eval("(t) => Number(t) > 0")(timeout_id0)
+        assert pm.eval("(t) => Number.isInteger(Number(t))")(timeout_id0)
         with pytest.raises(TypeError, match="going to be there"):
             await f1                                # `clearTimeout` not called
         f1 = loop.create_future()
@@ -37,8 +53,16 @@ def test_set_clear_timeout():
         # `setTimeout` should allow passing additional arguments to the callback, as spec-ed
         assert 3.0 == await pm.eval("new Promise((resolve) => setTimeout(function(){ resolve(arguments.length) }, 100, 90, 91, 92))")
         assert 92.0 == await pm.eval("new Promise((resolve) => setTimeout((...args) => { resolve(args[2]) }, 100, 90, 91, 92))")
-        # TODO (Tom Tang): test `setTimeout` setting delay to 0 if < 0
-        # TODO (Tom Tang): test `setTimeout` accepting string as the delay, coercing to a number like parseFloat
+        # test `setTimeout` setting delay to 0 if < 0
+        await asyncio.wait_for(pm.eval("new Promise((resolve) => setTimeout(resolve, 0))"), timeout=0.05)
+        await asyncio.wait_for(pm.eval("new Promise((resolve) => setTimeout(resolve, -10000))"), timeout=0.05) # won't be precisely 0s
+        # test `setTimeout` accepting string as the delay, coercing to a number.
+        # Number('100') -> 100, pass if the actual delay is > 90ms and < 150ms
+        await asyncio.wait_for(pm.eval("new Promise((resolve) => setTimeout(resolve, '100'))"), timeout=0.15) # won't be precisely 100ms
+        with pytest.raises(asyncio.exceptions.TimeoutError):
+            await asyncio.wait_for(pm.eval("new Promise((resolve) => setTimeout(resolve, '100'))"), timeout=0.09)
+        # Number("1 second") -> NaN -> delay turns to be 0s
+        await asyncio.wait_for(pm.eval("new Promise((resolve) => setTimeout(resolve, '1 second'))"), timeout=0.5) # won't be precisely 0s
 
         # passing an invalid ID to `clearTimeout` should silently do nothing; no exception is thrown.
         pm.eval("clearTimeout(NaN)")
@@ -198,7 +222,7 @@ def test_promises():
             # FIXME (Tom Tang): We currently handle Promise exceptions by converting the object thrown to a Python Exception object through `pyTypeFactory`
             #               <objects of this type are not handled by PythonMonkey yet>
             # await pm.eval("Promise.resolve().then(()=>{ throw {a:1,toString(){return'anything'}} })")
-        with pytest.raises(pm.SpiderMonkeyError, match="on line 1:\nTypeError: undefined has no properties"): # not going through the conversion
+        with pytest.raises(pm.SpiderMonkeyError, match="on line 1, column 31:\nTypeError: undefined has no properties"): # not going through the conversion
             await pm.eval("Promise.resolve().then(()=>{ (undefined).prop })")
 
         # TODO (Tom Tang): Modify this testcase once we support ES2020-style dynamic import
@@ -208,13 +232,13 @@ def test_promises():
         # TODO (Tom Tang): properly test unhandled rejection
 
         # await scheduled jobs on the Python event-loop
-        js_sleep = pm.eval("(second) => new Promise((resolve) => setTimeout(resolve, second*1000))")
-        def py_sleep(second): # asyncio.sleep has issues on Python 3.8
+        js_sleep = pm.eval("(seconds) => new Promise((resolve) => setTimeout(resolve, seconds*1000))")
+        def py_sleep(seconds): # asyncio.sleep has issues on Python 3.8
             loop = asyncio.get_running_loop()
             future = loop.create_future()
-            loop.call_later(second, lambda:future.set_result(None))
+            loop.call_later(seconds, lambda:future.set_result(None))
             return future
-        both_sleep = pm.eval("(js_sleep, py_sleep) => async (second) => { await js_sleep(second); await py_sleep(second) }")(js_sleep, py_sleep)
+        both_sleep = pm.eval("(js_sleep, py_sleep) => async (seconds) => { await js_sleep(seconds); await py_sleep(seconds) }")(js_sleep, py_sleep)
         await asyncio.wait_for(both_sleep(0.1), timeout=0.3) # won't be precisely 0.2s 
         with pytest.raises(asyncio.exceptions.TimeoutError):
             await asyncio.wait_for(both_sleep(0.1), timeout=0.19)
@@ -248,4 +272,4 @@ def test_webassembly():
 
         # making sure the async_fn is run
         return True
-    assert asyncio.run(async_fn())
+    assert asyncio.run(async_fn())
\ No newline at end of file
